from datetime import datetime, timedelta
from zoneinfo import ZoneInfo

import time_machine
from dateutil.relativedelta import relativedelta
from django.conf import settings
from django.test import TestCase, override_settings
from edc_protocol.research_protocol_config import ResearchProtocolConfig
from edc_utils import get_utcnow
from faker import Faker
from model_bakery import baker

from consent_app.models import SubjectConsent
from edc_consent.site_consents import site_consents

from ...consent_definition import ConsentDefinition
from ...exceptions import ConsentDefinitionDoesNotExist
from ..consent_test_utils import consent_factory

fake = Faker()


@time_machine.travel(datetime(2019, 4, 1, 8, 00, tzinfo=ZoneInfo("UTC")))
@override_settings(
    EDC_PROTOCOL_STUDY_OPEN_DATETIME=get_utcnow() - relativedelta(years=5),
    EDC_PROTOCOL_STUDY_CLOSE_DATETIME=get_utcnow() + relativedelta(years=1),
    EDC_AUTH_SKIP_SITE_AUTHS=True,
    EDC_AUTH_SKIP_AUTH_UPDATER=False,
)
class TestConsentModel(TestCase):
    def setUp(self):
        self.study_open_datetime = ResearchProtocolConfig().study_open_datetime
        self.study_close_datetime = ResearchProtocolConfig().study_close_datetime
        self.register_with_site_consents()
        self.dob = self.study_open_datetime - relativedelta(years=25)

        self.subject_identifier = "123456789"
        self.identity = "987654321"

    def register_with_site_consents(
        self,
        consent_v1: ConsentDefinition | None = None,
        consent_v2: ConsentDefinition | None = None,
        consent_v3: ConsentDefinition | None = None,
    ) -> tuple[ConsentDefinition, ConsentDefinition, ConsentDefinition]:
        site_consents.registry = {}
        if consent_v1 is None:
            consent_v1 = consent_factory(
                proxy_model="consent_app.subjectconsentv1",
                start=self.study_open_datetime,
                end=self.study_open_datetime + timedelta(days=50),
                version="1.0",
            )
        if consent_v2 is None:
            consent_v2 = consent_factory(
                proxy_model="consent_app.subjectconsentv2",
                start=self.study_open_datetime + timedelta(days=51),
                end=self.study_open_datetime + timedelta(days=100),
                version="2.0",
            )

        if consent_v3 is None:
            self.consent_v3_start_date = self.study_open_datetime + timedelta(days=101)
            consent_v3 = consent_factory(
                proxy_model="consent_app.subjectconsentv3",
                start=self.study_open_datetime + timedelta(days=101),
                end=self.study_open_datetime + timedelta(days=150),
                version="3.0",
                updates=consent_v2,
            )
        site_consents.register(consent_v1)
        site_consents.register(consent_v2, updated_by=consent_v3)
        site_consents.register(consent_v3)
        return consent_v1, consent_v2, consent_v3

    def create_v1_consent_for_subject(self) -> datetime:
        # travel to consent v1 validity period and consent subject
        traveller = time_machine.travel(self.study_open_datetime)
        traveller.start()
        consent_datetime = get_utcnow()
        cdef = site_consents.get_consent_definition(report_datetime=consent_datetime)
        baker.make_recipe(
            cdef.model,
            subject_identifier=self.subject_identifier,
            identity=self.identity,
            confirm_identity=self.identity,
            consent_datetime=consent_datetime,
            dob=get_utcnow() - relativedelta(years=25),
        )
        traveller.stop()
        return consent_datetime

    def create_v2_consent_for_subject(self) -> datetime:
        # travel to consent v2 validity period and consent subject
        traveller = time_machine.travel(self.study_open_datetime + timedelta(days=51))
        traveller.start()
        consent_datetime = get_utcnow()
        cdef = site_consents.get_consent_definition(report_datetime=consent_datetime)
        baker.make_recipe(
            cdef.model,
            subject_identifier=self.subject_identifier,
            identity=self.identity,
            confirm_identity=self.identity,
            consent_datetime=consent_datetime,
            dob=get_utcnow() - relativedelta(years=25),
        )
        traveller.stop()
        return consent_datetime

    def create_v3_consent_for_subject(self, days: int | None = None) -> datetime:
        """Use a consent datetime relative to the version 2 end date."""
        # travel to consent v3 validity period and consent subject
        days = days or 10
        cdef_v2 = site_consents.get_consent_definition(version="2.0")
        cdef = site_consents.get_consent_definition(report_datetime=cdef_v2.start)
        traveller = time_machine.travel(cdef.end + relativedelta(days=days))
        traveller.start()
        consent_datetime = get_utcnow()  # cdef.enf + xx days
        cdef = site_consents.get_consent_definition(report_datetime=consent_datetime)
        baker.make_recipe(
            cdef.model,
            subject_identifier=self.subject_identifier,
            identity=self.identity,
            confirm_identity=self.identity,
            consent_datetime=consent_datetime,
            dob=get_utcnow() - relativedelta(years=25),
        )
        traveller.stop()
        return consent_datetime

    def test_is_v2_within_v2_consent_period(self):
        consent_v1, consent_v2, consent_v3 = self.register_with_site_consents()
        self.create_v1_consent_for_subject()
        self.create_v2_consent_for_subject()
        self.create_v3_consent_for_subject()

        self.assertEqual(SubjectConsent.objects.filter(identity=self.identity).count(), 3)

        consent = site_consents.get_consent_or_raise(
            subject_identifier=self.subject_identifier,
            report_datetime=consent_v3.start - relativedelta(days=5),
            site_id=settings.SITE_ID,
        )
        self.assertEqual(consent.version, "2.0")

    def test_consent_date_is_for_version(self):
        """There should be no-gap! If they haven't signed V3 then
        data entry must stop unless there is overlap. In the case
        of overlap. the higher version consent should be used.
        """
        self.register_with_site_consents()
        self.create_v1_consent_for_subject()
        self.create_v2_consent_for_subject()
        v3_consent_datetime = self.create_v3_consent_for_subject(days=10)

        self.assertEqual(SubjectConsent.objects.filter(identity=self.identity).count(), 3)
        cdef = site_consents.get_consent_definition(report_datetime=v3_consent_datetime)
        cosent_obj_v3 = SubjectConsent.objects.get(
            consent_datetime__range=[cdef.start, cdef.end]
        )
        consent = site_consents.get_consent_or_raise(
            subject_identifier=self.subject_identifier,
            report_datetime=cosent_obj_v3.consent_datetime - relativedelta(days=6),
            site_id=settings.SITE_ID,
        )
        self.assertEqual(consent.version, "3.0")
        consent = site_consents.get_consent_or_raise(
            subject_identifier=self.subject_identifier,
            report_datetime=cosent_obj_v3.consent_datetime - relativedelta(days=11),
            site_id=settings.SITE_ID,
        )
        self.assertEqual(consent.version, "2.0")

    def test_v3_consent_date_gap(self):
        """Assert raises if no consent definition covers the intended
        consent date.
        """
        consent_v2 = consent_factory(
            proxy_model="consent_app.subjectconsentv2",
            start=self.study_open_datetime + timedelta(days=51),
            end=self.study_open_datetime + timedelta(days=100),
            version="2.0",
        )
        consent_v3 = consent_factory(
            proxy_model="consent_app.subjectconsentv3",
            start=self.study_open_datetime + timedelta(days=120),
            end=self.study_open_datetime + timedelta(days=150),
            version="3.0",
            updates=consent_v2,
        )
        self.register_with_site_consents(consent_v2=consent_v2, consent_v3=consent_v3)

        self.create_v1_consent_for_subject()
        self.create_v2_consent_for_subject()
        # cannot consent, date does not fall within a consent period
        self.assertRaises(
            ConsentDefinitionDoesNotExist, self.create_v3_consent_for_subject, days=10
        )

        # ok, date does falls within a consent period
        self.create_v3_consent_for_subject(days=30)

    def test_is_v3_on_v3_consent_date(self):
        self.register_with_site_consents()
        self.create_v1_consent_for_subject()
        self.create_v2_consent_for_subject()
        v3_consent_datetime = self.create_v3_consent_for_subject()

        self.assertEqual(SubjectConsent.objects.filter(identity=self.identity).count(), 3)

        consent = site_consents.get_consent_or_raise(
            subject_identifier=self.subject_identifier,
            report_datetime=v3_consent_datetime,
            site_id=settings.SITE_ID,
        )
        self.assertEqual(consent.version, "3.0")

    def test_is_v3_on_after_v3_consent_date(self):
        self.register_with_site_consents()
        self.create_v1_consent_for_subject()
        self.create_v2_consent_for_subject()
        v3_consent_datetime = self.create_v3_consent_for_subject()

        self.assertEqual(SubjectConsent.objects.filter(identity=self.identity).count(), 3)

        consent = site_consents.get_consent_or_raise(
            subject_identifier=self.subject_identifier,
            report_datetime=v3_consent_datetime + relativedelta(days=5),
            site_id=settings.SITE_ID,
        )
        self.assertEqual(consent.version, "3.0")



from django.db import models
from django.db.models import PROTECT, Manager
from edc_constants.choices import GENDER_UNDETERMINED
from edc_constants.constants import FEMALE
from edc_identifier.model_mixins import NonUniqueSubjectIdentifierModelMixin
from edc_model.models import BaseUuidModel, HistoricalRecords
from edc_registration.model_mixins import UpdatesOrCreatesRegistrationModelMixin
from edc_sites.managers import CurrentSiteManager
from edc_sites.model_mixins import SiteModelMixin
from edc_utils import get_utcnow
from edc_visit_schedule.model_mixins import VisitScheduleModelMixin

from edc_consent.field_mixins import (
    CitizenFieldsMixin,
    IdentityFieldsMixin,
    PersonalFieldsMixin,
    ReviewFieldsMixin,
    VulnerabilityFieldsMixin,
)
from edc_consent.managers import ConsentObjectsByCdefManager, CurrentSiteByCdefManager
from edc_consent.model_mixins import (
    ConsentExtensionModelMixin,
    ConsentModelMixin,
    RequiresConsentFieldsModelMixin,
)


class SubjectScreening(SiteModelMixin, BaseUuidModel):

    screening_identifier = models.CharField(max_length=25, unique=True)

    initials = models.CharField(max_length=5, default="TO")

    age_in_years = models.IntegerField(default=25)

    gender = models.CharField(
        max_length=5,
        choices=GENDER_UNDETERMINED,
        default=FEMALE,
    )

    report_datetime = models.DateTimeField()

    eligible = models.BooleanField(default=False)

    eligibility_datetime = models.DateTimeField()

    objects = models.Manager()
    on_site = CurrentSiteManager()
    history = HistoricalRecords()

    class Meta:
        pass


class SubjectConsent(
    ConsentModelMixin,
    SiteModelMixin,
    NonUniqueSubjectIdentifierModelMixin,
    UpdatesOrCreatesRegistrationModelMixin,
    IdentityFieldsMixin,
    ReviewFieldsMixin,
    PersonalFieldsMixin,
    CitizenFieldsMixin,
    VulnerabilityFieldsMixin,
    BaseUuidModel,
):
    screening_identifier = models.CharField(
        verbose_name="Screening identifier", max_length=50, unique=True
    )
    history = HistoricalRecords()

    class Meta(ConsentModelMixin.Meta):
        pass


class SubjectConsentV1(SubjectConsent):
    on_site = CurrentSiteByCdefManager()
    objects = ConsentObjectsByCdefManager()

    class Meta:
        proxy = True


class SubjectConsentV1Ext(ConsentExtensionModelMixin, SiteModelMixin, BaseUuidModel):

    subject_consent = models.ForeignKey(SubjectConsentV1, on_delete=models.PROTECT)

    on_site = CurrentSiteManager()
    history = HistoricalRecords()
    objects = Manager()

    class Meta(ConsentExtensionModelMixin.Meta, BaseUuidModel.Meta):
        verbose_name = "Subject Consent Extension V1.1"
        verbose_name_plural = "Subject Consent Extension V1.1"


class SubjectConsentUgV1(SubjectConsent):
    on_site = CurrentSiteByCdefManager()
    objects = ConsentObjectsByCdefManager()

    class Meta:
        proxy = True


class SubjectConsentV2(SubjectConsent):
    on_site = CurrentSiteByCdefManager()
    objects = ConsentObjectsByCdefManager()

    class Meta:
        proxy = True


class SubjectConsentV3(SubjectConsent):
    on_site = CurrentSiteByCdefManager()
    objects = ConsentObjectsByCdefManager()

    class Meta:
        proxy = True


class SubjectConsentV4(SubjectConsent):
    on_site = CurrentSiteByCdefManager()
    objects = ConsentObjectsByCdefManager()

    class Meta:
        proxy = True


class SubjectConsentUpdateToV3(SubjectConsent):
    class Meta:
        proxy = True


class SubjectReconsent(
    ConsentModelMixin,
    SiteModelMixin,
    NonUniqueSubjectIdentifierModelMixin,
    UpdatesOrCreatesRegistrationModelMixin,
    IdentityFieldsMixin,
    ReviewFieldsMixin,
    PersonalFieldsMixin,
    CitizenFieldsMixin,
    VulnerabilityFieldsMixin,
    BaseUuidModel,
):
    screening_identifier = models.CharField(
        verbose_name="Screening identifier", max_length=50, unique=True
    )
    history = HistoricalRecords()

    class Meta(ConsentModelMixin.Meta):
        pass


class SubjectConsent2(
    ConsentModelMixin,
    SiteModelMixin,
    NonUniqueSubjectIdentifierModelMixin,
    UpdatesOrCreatesRegistrationModelMixin,
    IdentityFieldsMixin,
    ReviewFieldsMixin,
    PersonalFieldsMixin,
    CitizenFieldsMixin,
    VulnerabilityFieldsMixin,
    BaseUuidModel,
):
    screening_identifier = models.CharField(
        verbose_name="Screening identifier", max_length=50, unique=True
    )

    history = HistoricalRecords()

    class Meta(ConsentModelMixin.Meta):
        pass


class SubjectVisit(
    SiteModelMixin, RequiresConsentFieldsModelMixin, VisitScheduleModelMixin, BaseUuidModel
):
    subject_identifier = models.CharField(max_length=25)
    report_datetime = models.DateTimeField(default=get_utcnow)

    # appointment = models.OneToOneField(Appointment, on_delete=CASCADE)
    # history = HistoricalRecords()
    pass


class TestModel(
    NonUniqueSubjectIdentifierModelMixin,
    RequiresConsentFieldsModelMixin,
    SiteModelMixin,
    BaseUuidModel,
):
    report_datetime = models.DateTimeField(default=get_utcnow)


class CrfOne(
    RequiresConsentFieldsModelMixin,
    NonUniqueSubjectIdentifierModelMixin,
    SiteModelMixin,
    BaseUuidModel,
):
    subject_visit = models.ForeignKey(SubjectVisit, on_delete=PROTECT)

    report_datetime = models.DateTimeField(default=get_utcnow)

    @property
    def related_visit(self):
        return self.subject_visit



from __future__ import annotations

from dataclasses import KW_ONLY, dataclass, field
from datetime import datetime
from typing import TYPE_CHECKING, Type

from django.apps import apps as django_apps
from django.core.exceptions import ObjectDoesNotExist
from edc_constants.constants import YES
from edc_protocol.research_protocol_config import ResearchProtocolConfig
from edc_sites import site_sites
from edc_utils import formatted_date
from edc_utils.date import to_local
from edc_visit_schedule.schedule import VisitCollection

from .exceptions import ConsentDefinitionError

if TYPE_CHECKING:
    from edc_identifier.model_mixins import UniqueSubjectIdentifierModelMixin
    from edc_model.models import BaseUuidModel
    from edc_sites.model_mixins import SiteModelMixin

    from .consent_definition import ConsentDefinition

    class ConsentLikeModel(SiteModelMixin, UniqueSubjectIdentifierModelMixin, BaseUuidModel):
        _meta: ...

    class ConsentExtensionLikeModel(
        SiteModelMixin, UniqueSubjectIdentifierModelMixin, BaseUuidModel
    ):
        agrees_to_extension: str = field(default=YES, compare=False)
        _meta: ...


@dataclass(order=True)
class ConsentDefinitionExtension:
    """A definition to truncate the number of visits/timepoints in a
    visit collection for a consented subject, if necessary.

    If the consent extension model is complete for this subject and
    the field `agrees_to_extension` == YES, the visit collection
    is NOT truncated.

    For example, a trial originally consents to a 36m followup. At some
    point the trial receives approval to extend followup to 48m for
    those who agree. For those who DO NOT agree to the extended
    followup, the timepoints defined in this extension are removed
    from the given visit collection.

    Note: See also `Schedule`. The schedule should be defined with
    all possible visits/timepoints. That is, if approval is for 48m
    of followup, the Schedule should be defined with 48m of followup.
    This class will remove, not add, visits/timepoints from the given
    visit collection if necessary.
    """

    model: str = field(compare=False)
    _ = KW_ONLY
    start: datetime = field(default=ResearchProtocolConfig().study_open_datetime, compare=True)
    version: str = field(default="1", compare=False)
    extends: ConsentDefinition = field(default=None, compare=False)
    timepoints: list[int] | None = field(default_factory=list, compare=False)
    site_ids: list[int] = field(default_factory=list, compare=False)
    country: str | None = field(default=None, compare=False)

    name: str = field(init=False, compare=False)
    sort_index: str = field(init=False)

    def __post_init__(self):
        self.name = f"{self.model}-{self.version}"
        self.sort_index = self.name
        if not self.start.tzinfo:
            raise ConsentDefinitionError(f"Naive datetime not allowed. Got {self.start}.")
        elif str(self.start.tzinfo) != "UTC":
            raise ConsentDefinitionError(f"Start date must be UTC. Got {self.start}.")
        self.extends.check_date_within_study_period()

    def update_visit_collection(
        self,
        visits: VisitCollection = None,
        subject_identifier: str = None,
        site_id: str = None,
        original_visit_collection: VisitCollection = None,
    ) -> VisitCollection:
        """Returns the visit collection with or without the
        timepoints of this extension.

        `original_visit_collection` is unchanged.
        """
        if not self.get_consent_extension_for(
            subject_identifier=subject_identifier,
            site_id=site_id,
        ):
            for v in original_visit_collection.values():
                if v.timepoint in self.timepoints:
                    del visits[v.code]
        return visits

    @property
    def model_cls(self) -> Type[ConsentExtensionLikeModel]:
        return django_apps.get_model(self.model)

    @property
    def sites(self):
        if not site_sites.loaded:
            raise ConsentDefinitionError(
                "No registered sites found or edc_sites.sites not loaded yet. "
                "Perhaps place `edc_sites` before `edc_consent` "
                "in INSTALLED_APPS."
            )
        if self.country:
            sites = site_sites.get_by_country(self.country, aslist=True)
        elif self.site_ids:
            sites = [s for s in site_sites.all(aslist=True) if s.site_id in self.site_ids]
        else:
            sites = [s for s in site_sites.all(aslist=True)]
        return sites

    def get_consent_extension_for(self, **kwargs) -> ConsentExtensionLikeModel | None:
        """Returns the consent extension model instance for the
        parent consent definition.

        If field `agrees_to_extension` == YES, extension is granted.
        """
        subject_consent = self.get_consent_for(**kwargs)
        try:
            consent_extension_obj = self.model_cls.objects.get(
                subject_consent=subject_consent,
                report_datetime__gte=self.start,
                agrees_to_extension=YES,
            )
        except ObjectDoesNotExist:
            consent_extension_obj = None
        return consent_extension_obj

    def get_consent_for(self, **kwargs) -> ConsentLikeModel | None:
        """Returns the parent consent model instance for the subject."""
        return self.extends.get_consent_for(**kwargs)

    @property
    def display_name(self) -> str:
        return (
            f"{self.model_cls._meta.verbose_name} v{self.version} valid "
            f"from {formatted_date(to_local(self.start))} to "
            f"{formatted_date(to_local(self.extends.end))}"
        )

    @property
    def verbose_name(self) -> str:
        return self.model_cls._meta.verbose_name



from __future__ import annotations

from datetime import datetime
from typing import TYPE_CHECKING

from django.utils.translation import gettext as _
from edc_form_validators import INVALID_ERROR
from edc_sites.site import sites as site_sites

from edc_consent.consent_definition import ConsentDefinition
from edc_consent.exceptions import (
    ConsentDefinitionDoesNotExist,
    ConsentDefinitionNotConfiguredForUpdate,
    NotConsentedError,
    SiteConsentError,
)
from edc_consent.site_consents import site_consents

if TYPE_CHECKING:
    from django.contrib.sites.models import Site


class ConsentDefinitionFormValidatorMixin:

    @property
    def subject_consent(self):
        cdef = self.get_consent_definition()
        return cdef.model_cls.objects.get(subject_identifier=self.subject_identifier)

    def get_consent_datetime_or_raise(
        self,
        report_datetime: datetime = None,
        site: Site = None,
        fldname: str = None,
        error_code: str = None,
    ) -> datetime:
        """Returns the consent_datetime of this subject"""
        consent_obj = self.get_consent_or_raise(
            report_datetime=report_datetime, site=site, fldname=fldname, error_code=error_code
        )
        return consent_obj.consent_datetime

    def get_consent_or_raise(
        self,
        report_datetime: datetime = None,
        site: Site = None,
        fldname: str | None = None,
        error_code: str | None = None,
    ) -> datetime:
        """Returns the consent_datetime of this subject.

        Wraps func `consent_datetime_or_raise` to re-raise exceptions
        as ValidationError.
        """

        fldname = fldname or "report_datetime"
        error_code = error_code or INVALID_ERROR
        try:
            consent_obj = site_consents.get_consent_or_raise(
                subject_identifier=self.subject_identifier,
                report_datetime=report_datetime,
                site_id=getattr(site, "id", None),
            )
        except (NotConsentedError, ConsentDefinitionNotConfiguredForUpdate) as e:
            self.raise_validation_error({fldname: str(e)}, error_code)
        return consent_obj

    def get_consent_definition(
        self, report_datetime: datetime = None, fldname: str = None, error_code: str = None
    ) -> ConsentDefinition:
        # get the consent definition (must be from this schedule)
        schedule = getattr(self, "related_visit", self.instance).schedule

        site = getattr(self, "related_visit", self.instance).site
        try:
            consent_definition = schedule.get_consent_definition(
                site=site_sites.get(site.id),
                report_datetime=report_datetime,
            )
        except ConsentDefinitionDoesNotExist as e:
            self.raise_validation_error({fldname: str(e)}, error_code)
        except SiteConsentError:
            possible_consents = "', '".join(
                [cdef.display_name for cdef in site_consents.consent_definitions]
            )
            self.raise_validation_error(
                {
                    fldname: _(
                        "Date does not fall within a valid consent period. "
                        "Possible consents are '%(possible_consents)s'. "
                        % {"possible_consents": possible_consents}
                    )
                },
                error_code,
            )
        return consent_definition



from __future__ import annotations

from django.db import models
from edc_search.model_mixins import SearchSlugManager
from edc_sites.managers import CurrentSiteManager

from .site_consents import site_consents


class ConsentObjectsManager(SearchSlugManager, models.Manager):
    use_in_migrations = True

    def get_by_natural_key(self, subject_identifier_as_pk):
        return self.get(subject_identifier_as_pk=subject_identifier_as_pk)


class ConsentObjectsByCdefManager(ConsentObjectsManager):
    """An objects model manager to use on consent proxy models
    linked to a ConsentDefinition.

    Filters queryset by the proxy model's label_lower
    """

    def get_queryset(self):
        qs = super().get_queryset()
        cdef = site_consents.get_consent_definition(model=qs.model._meta.label_lower)
        return qs.filter(version=cdef.version)


class CurrentSiteByCdefManager(CurrentSiteManager):
    """A site model manager to use on consent proxy models
    linked to a ConsentDefinition.

    Filters queryset by the proxy model's label_lower
    """

    def get_queryset(self):
        qs = super().get_queryset()
        cdef = site_consents.get_consent_definition(model=qs.model._meta.label_lower)
        return qs.filter(site_id=cdef.site.site_id, version=cdef.version)



from __future__ import annotations

from typing import TYPE_CHECKING

from dateutil.relativedelta import relativedelta
from edc_visit_schedule.schedule import Schedule
from edc_visit_schedule.visit import Crf, CrfCollection, Visit
from edc_visit_schedule.visit_schedule import VisitSchedule

if TYPE_CHECKING:
    from edc_consent.consent_definition import ConsentDefinition


def get_visit_schedule(
    consent_definition: ConsentDefinition | list[ConsentDefinition], extend: bool | None = None
) -> VisitSchedule:
    crfs = CrfCollection(Crf(show_order=1, model="consent_app.crfone", required=True))

    visit = Visit(
        code="1000",
        timepoint=0,
        rbase=relativedelta(days=0),
        rlower=relativedelta(days=0),
        rupper=relativedelta(days=6),
        requisitions=None,
        crfs=crfs,
        requisitions_unscheduled=None,
        crfs_unscheduled=None,
        allow_unscheduled=False,
        facility_name="5-day-clinic",
    )

    visit1010 = Visit(
        code="1010",
        timepoint=1,
        rbase=relativedelta(months=1),
        rlower=relativedelta(days=0),
        rupper=relativedelta(days=6),
        requisitions=None,
        crfs=crfs,
        requisitions_unscheduled=None,
        crfs_unscheduled=None,
        allow_unscheduled=False,
        facility_name="5-day-clinic",
    )

    visit1020 = Visit(
        code="1020",
        timepoint=2,
        rbase=relativedelta(months=2),
        rlower=relativedelta(days=0),
        rupper=relativedelta(days=6),
        requisitions=None,
        crfs=crfs,
        requisitions_unscheduled=None,
        crfs_unscheduled=None,
        allow_unscheduled=False,
        facility_name="5-day-clinic",
    )

    schedule = Schedule(
        name="schedule1",
        onschedule_model="edc_appointment.onscheduleone",
        offschedule_model="edc_appointment.offscheduleone",
        appointment_model="edc_appointment.appointment",
        consent_definitions=consent_definition,
    )

    visit_schedule = VisitSchedule(
        name="visit_schedule",
        offstudy_model="edc_appointment.subjectoffstudy",
        death_report_model="edc_appointment.deathreport",
        locator_model="edc_locator.subjectlocator",
    )

    schedule.add_visit(visit)
    if extend:
        schedule.add_visit(visit1010)
        schedule.add_visit(visit1020)

    visit_schedule.add_schedule(schedule)
    return visit_schedule



from django.db.models.signals import pre_save
from django.dispatch import receiver
from edc_sites import site_sites

from ..model_mixins import RequiresConsentFieldsModelMixin
from ..site_consents import site_consents


@receiver(pre_save, weak=False, dispatch_uid="requires_consent_on_pre_save")
def requires_consent_on_pre_save(instance, raw, using, update_fields, **kwargs):
    if (
        not raw
        and not update_fields
        and isinstance(instance, (RequiresConsentFieldsModelMixin,))
        and not instance._meta.model_name.startswith("historical")
    ):
        subject_identifier = getattr(instance, "related_visit", instance).subject_identifier
        site = getattr(instance, "related_visit", instance).site
        try:
            schedule = getattr(instance, "related_visit", instance).schedule
        except AttributeError:
            schedule = None
        if schedule:
            consent_definition = schedule.get_consent_definition(
                site=site_sites.get(site.id), report_datetime=instance.report_datetime
            )
        else:
            # this is a PRN model, like SubjectLocator, with no visit_schedule
            consent_definition = site_consents.get_consent_definition(
                site=site_sites.get(site.id), report_datetime=instance.report_datetime
            )
        site_consents.get_consent_or_raise(
            subject_identifier=subject_identifier,
            report_datetime=instance.report_datetime,
            site_id=site.id,
        )
        version = consent_definition.version
        if (
            consent_definition.extended_by
            and consent_definition.extended_by.start <= instance.report_datetime
            and consent_definition.extended_by.get_consent_extension_for(
                subject_identifier=subject_identifier,
                site_id=instance.site_id,
            )
        ):
            version = consent_definition.extended_by.version
        instance.consent_version = version
        instance.consent_model = consent_definition.model



from .consent_extension_model_mixin import ConsentExtensionModelMixin
from .consent_model_mixin import ConsentModelMixin
from .consent_version_model_mixin import ConsentVersionModelMixin
from .requires_consent_fields_model_mixin import RequiresConsentFieldsModelMixin

__all__ = [
    "ConsentModelMixin",
    "RequiresConsentFieldsModelMixin",
    "ConsentVersionModelMixin",
    "ConsentExtensionModelMixin",
]



from django.urls import path

from .admin_site import edc_consent_admin
from .views import HomeView

app_name = "edc_consent"

urlpatterns = [
    path("/admin/", edc_consent_admin.urls),
    path("", HomeView.as_view(), name="home_url"),
]



from datetime import datetime, timedelta
from zoneinfo import ZoneInfo

import time_machine
from dateutil.relativedelta import relativedelta
from django.conf import settings
from django.test import TestCase, override_settings
from edc_protocol.research_protocol_config import ResearchProtocolConfig
from edc_utils import get_utcnow
from model_bakery import baker

from edc_consent import site_consents
from edc_consent.consent_definition_extension import ConsentDefinitionExtension
from edc_consent.tests.consent_test_utils import consent_factory


class TestConsentExtension(TestCase):
    @time_machine.travel(datetime(2019, 4, 1, 8, 00, tzinfo=ZoneInfo("UTC")))
    @override_settings(
        EDC_PROTOCOL_STUDY_OPEN_DATETIME=get_utcnow() - relativedelta(years=5),
        EDC_PROTOCOL_STUDY_CLOSE_DATETIME=get_utcnow() + relativedelta(years=1),
        EDC_AUTH_SKIP_SITE_AUTHS=True,
        EDC_AUTH_SKIP_AUTH_UPDATER=False,
    )
    class TestConsentModel(TestCase):
        def setUp(self):
            self.study_open_datetime = ResearchProtocolConfig().study_open_datetime
            self.study_close_datetime = ResearchProtocolConfig().study_close_datetime
            site_consents.registry = {}
            self.consent_v1 = consent_factory(
                proxy_model="consent_app.subjectconsentv1",
                start=self.study_open_datetime,
                end=self.study_open_datetime + timedelta(days=50),
                version="1.0",
            )
            self.consent_v1_ext = ConsentDefinitionExtension(
                "consent_app.subjectconsentv11",
                version="1.1",
                start=datetime(2024, 12, 16, tzinfo=ZoneInfo("UTC")),
                extends=self.consent_v1,
                timepoints=list(range(15, 18 + 1)),
            )
            site_consents.register(self.consent_v1, extended_by=self.consent_v1_ext)
            self.dob = self.study_open_datetime - relativedelta(years=25)

        def test_consent_version_extension(self):
            subject_identifier = "123456789"
            identity = "987654321"
            baker.make_recipe(
                "consent_app.subjectconsentv1",
                subject_identifier=subject_identifier,
                identity=identity,
                confirm_identity=identity,
                consent_datetime=self.study_open_datetime + timedelta(days=1),
                dob=get_utcnow() - relativedelta(years=25),
            )
            subject_consent = site_consents.get_consent_or_raise(
                subject_identifier="123456789",
                report_datetime=self.study_open_datetime + timedelta(days=1),
                site_id=settings.SITE_ID,
            )
            self.assertEqual(subject_consent.version, "1.0")
            baker.make_recipe(
                "consent_app.subjectconsentv1ext",
                subject_identifier=subject_identifier,
                identity=identity,
                confirm_identity=identity,
                consent_datetime=self.study_open_datetime + timedelta(days=1),
                dob=get_utcnow() - relativedelta(years=25),
            )



import string
from secrets import choice

from dateutil.relativedelta import relativedelta
from django.contrib.auth.models import User
from django.http.request import HttpRequest
from django.test import TestCase, override_settings
from edc_protocol.research_protocol_config import ResearchProtocolConfig
from edc_utils import get_utcnow
from faker import Faker
from model_bakery import baker

from consent_app.models import SubjectConsentV1
from edc_consent.actions import unverify_consent, verify_consent
from edc_consent.site_consents import site_consents

from ..consent_test_utils import consent_definition_factory

fake = Faker()


@override_settings(
    EDC_PROTOCOL_STUDY_OPEN_DATETIME=get_utcnow() - relativedelta(years=5),
    EDC_PROTOCOL_STUDY_CLOSE_DATETIME=get_utcnow() + relativedelta(years=1),
    EDC_AUTH_SKIP_SITE_AUTHS=True,
    EDC_AUTH_SKIP_AUTH_UPDATER=False,
)
class TestActions(TestCase):
    def setUp(self):
        super().setUp()
        site_consents.registry = {}
        self.study_open_datetime = ResearchProtocolConfig().study_open_datetime
        self.study_close_datetime = ResearchProtocolConfig().study_close_datetime
        cdef = consent_definition_factory(
            start=self.study_open_datetime, end=self.study_close_datetime
        )
        site_consents.register(cdef)
        self.request = HttpRequest()
        user = User.objects.create(username="erikvw")
        self.request.user = user
        for _ in range(3):
            first_name = fake.first_name()
            last_name = fake.last_name()
            initials = first_name[0] + choice(string.ascii_uppercase) + last_name[0]
            baker.make_recipe(
                "consent_app.subjectconsentv1",
                consent_datetime=self.study_open_datetime + relativedelta(days=1),
                initials=initials.upper(),
            )

    def test_verify(self):
        for consent_obj in SubjectConsentV1.objects.all():
            verify_consent(request=self.request, consent_obj=consent_obj)
        for consent_obj in SubjectConsentV1.objects.all():
            self.assertTrue(consent_obj.is_verified)
            self.assertEqual(consent_obj.verified_by, "erikvw")
            self.assertIsNotNone(consent_obj.is_verified_datetime)

    def test_unverify(self):
        for consent_obj in SubjectConsentV1.objects.all():
            unverify_consent(consent_obj=consent_obj)
        for consent_obj in SubjectConsentV1.objects.all():
            self.assertFalse(consent_obj.is_verified)
            self.assertIsNone(consent_obj.verified_by)
            self.assertIsNone(consent_obj.is_verified_datetime)



from __future__ import annotations

from dataclasses import KW_ONLY, dataclass, field
from datetime import datetime
from typing import TYPE_CHECKING, Type

from django.apps import apps as django_apps
from django.core.exceptions import ObjectDoesNotExist
from edc_constants.constants import FEMALE, MALE
from edc_protocol.research_protocol_config import ResearchProtocolConfig
from edc_screening.utils import get_subject_screening_model
from edc_sites import site_sites
from edc_utils import ceil_secs, floor_secs, formatted_date, formatted_datetime
from edc_utils.date import to_local

from .exceptions import (
    ConsentDefinitionError,
    ConsentDefinitionValidityPeriodError,
    NotConsentedError,
)

if TYPE_CHECKING:
    from edc_model.models import BaseUuidModel
    from edc_screening.model_mixins import EligibilityModelMixin, ScreeningModelMixin

    from .consent_definition_extension import ConsentDefinitionExtension
    from .stubs import ConsentLikeModel

    class SubjectScreening(ScreeningModelMixin, EligibilityModelMixin, BaseUuidModel): ...


@dataclass(order=True)
class ConsentDefinition:
    """A class that represents the general attributes
    of a consent.
    """

    proxy_model: str = field(compare=False)
    _ = KW_ONLY
    start: datetime = field(default=ResearchProtocolConfig().study_open_datetime, compare=True)
    end: datetime = field(default=ResearchProtocolConfig().study_close_datetime, compare=False)
    version: str = field(default="1", compare=False)
    updates: ConsentDefinition = field(default=None, compare=False)
    extends: ConsentDefinition = field(default=None, compare=False)
    screening_model: list[str] = field(default_factory=list, compare=False)
    age_min: int = field(default=18, compare=False)
    age_max: int = field(default=110, compare=False)
    age_is_adult: int = field(default=18, compare=False)
    gender: list[str] | None = field(default_factory=list, compare=False)
    site_ids: list[int] = field(default_factory=list, compare=False)
    country: str | None = field(default=None, compare=False)
    validate_duration_overlap_by_model: bool | None = field(default=True, compare=False)
    subject_type: str = field(default="subject", compare=False)
    timepoints: list[int] | None = field(default_factory=list, compare=False)

    name: str = field(init=False, compare=False)
    # set updated_by when the cdef is registered, see site_consents
    updated_by: ConsentDefinition = field(default=None, compare=False, init=False)
    extended_by: ConsentDefinitionExtension = field(default=None, compare=False, init=False)
    _model: str = field(init=False, compare=False)
    sort_index: str = field(init=False)

    def __post_init__(self):
        self.model = self.proxy_model
        self.name = f"{self.proxy_model}-{self.version}"
        self.sort_index = self.name
        self.gender = [MALE, FEMALE] if not self.gender else self.gender
        if not self.screening_model:
            self.screening_model = [get_subject_screening_model()]
        if MALE not in self.gender and FEMALE not in self.gender:
            raise ConsentDefinitionError(f"Invalid gender. Got {self.gender}.")
        if not self.start.tzinfo:
            raise ConsentDefinitionError(f"Naive datetime not allowed. Got {self.start}.")
        elif str(self.start.tzinfo).upper() != "UTC":
            raise ConsentDefinitionError(
                f"Start date must be UTC. Got {self.start} / {self.start.tzinfo}."
            )
        if not self.end.tzinfo:
            raise ConsentDefinitionError(f"Naive datetime not allowed Got {self.end}.")
        elif str(self.end.tzinfo).upper() != "UTC":
            raise ConsentDefinitionError(
                f"End date must be UTC. Got {self.end} / {self.start.tzinfo}."
            )
        self.check_date_within_study_period()

    @property
    def model(self):
        from .managers import ConsentObjectsByCdefManager, CurrentSiteByCdefManager

        model_cls = django_apps.get_model(self._model)
        if not model_cls._meta.proxy:
            raise ConsentDefinitionError(
                f"Model class must be a proxy. See {self.name}. Got {model_cls}"
            )
        elif not isinstance(model_cls.objects, (ConsentObjectsByCdefManager,)):
            raise ConsentDefinitionError(
                "Incorrect 'objects' model manager for consent model. "
                f"Expected {ConsentObjectsByCdefManager}. See {self.name}.  "
                f"Got {model_cls.objects.__class__}"
            )
        elif not isinstance(model_cls.on_site, (CurrentSiteByCdefManager,)):
            raise ConsentDefinitionError(
                "Incorrect 'on_site' model manager for consent model. "
                f"Expected {CurrentSiteByCdefManager}. See {self.name}.  "
                f"Got {model_cls.objects.__class__}"
            )
        return self._model

    @model.setter
    def model(self, value):
        self._model = value

    @property
    def sites(self):
        if not site_sites.loaded:
            raise ConsentDefinitionError(
                "No registered sites found or edc_sites.sites not loaded yet. "
                "Perhaps place `edc_sites` before `edc_consent` "
                "in INSTALLED_APPS."
            )
        if self.country:
            sites = site_sites.get_by_country(self.country, aslist=True)
        elif self.site_ids:
            sites = [s for s in site_sites.all(aslist=True) if s.site_id in self.site_ids]
        else:
            sites = [s for s in site_sites.all(aslist=True)]
        return sites

    def get_consent_for(
        self,
        subject_identifier: str = None,
        site_id: int | None = None,
        raise_if_not_consented: bool | None = None,
    ) -> ConsentLikeModel | None:
        raise_if_not_consented = (
            True if raise_if_not_consented is None else raise_if_not_consented
        )
        opts: dict[str, str | int] = dict(
            subject_identifier=subject_identifier, version=self.version
        )
        if site_id:
            opts.update(site_id=site_id)
        try:
            consent_obj = self.model_cls.objects.get(**opts)
        except ObjectDoesNotExist:
            consent_obj = None
        if not consent_obj and raise_if_not_consented:
            raise NotConsentedError(
                f"Consent not found for this version. Has subject '{subject_identifier}' "
                f"completed a version '{self.version}' consent?"
            )
        return consent_obj

    @property
    def model_cls(self) -> Type[ConsentLikeModel]:
        return django_apps.get_model(self.model)

    @property
    def display_name(self) -> str:
        return (
            f"{self.model_cls._meta.verbose_name} v{self.version} valid "
            f"from {formatted_date(to_local(self.start))} to "
            f"{formatted_date(to_local(self.end))}"
        )

    @property
    def verbose_name(self) -> str:
        return self.model_cls._meta.verbose_name

    def valid_for_datetime_or_raise(self, report_datetime: datetime) -> None:
        if report_datetime and not (
            floor_secs(self.start) <= report_datetime <= ceil_secs(self.end)
        ):
            date_string = formatted_date(report_datetime)
            raise ConsentDefinitionValidityPeriodError(
                "Date does not fall within the validity period."
                f"See {self.name}. Got {date_string}. "
            )

    def check_date_within_study_period(self) -> None:
        """Raises if the date is not within the opening and closing
        dates of the protocol.
        """
        protocol = ResearchProtocolConfig()
        study_open_datetime = protocol.study_open_datetime
        study_close_datetime = protocol.study_close_datetime
        for attr in ["start", "end"]:
            if not (
                floor_secs(study_open_datetime)
                <= getattr(self, attr)
                <= ceil_secs(study_close_datetime)
            ):
                open_date_string = formatted_datetime(study_open_datetime)
                close_date_string = formatted_datetime(study_close_datetime)
                attr_date_string = formatted_datetime(getattr(self, attr))
                raise ConsentDefinitionError(
                    f"Invalid {attr} date. "
                    f"Must be within the opening and closing dates of the protocol. "
                    f"See {self}. "
                    f"Got {open_date_string=}, {close_date_string=}, "
                    f"{attr=}, {attr_date_string=}."
                )

    def get_previous_consent(
        self, subject_identifier: str, exclude_id=None
    ) -> ConsentLikeModel:
        previous_consent = (
            self.model_cls.objects.filter(subject_identifier=subject_identifier)
            .exclude(id=exclude_id)
            .order_by("consent_datetime")
        )
        if previous_consent.count() > 0:
            return previous_consent.last()
        else:
            raise ObjectDoesNotExist("Previous consent does not exist")



from django.db import models


class SiteFieldsMixin(models.Model):
    site_code = models.CharField(
        verbose_name="Site",
        max_length=25,
        help_text=(
            "This refers to the site or 'clinic area' where the " "subject is being consented."
        ),
    )

    class Meta:
        abstract = True
